@@ -5,13 +5,14 @@
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 from __future__ import annotations
 
-from typing import Any, Awaitable, Iterable
+from typing import Any, Awaitable, Iterable, TypeVar
 from urllib.parse import quote as urllib_quote
 
 from mautrix.api import Method, Path
 from mautrix.client import ClientAPI, StoreUpdatingAPI
 from mautrix.errors import (
     IntentError,
+    MAlreadyJoined,
     MatrixRequestError,
     MBadState,
     MForbidden,
@@ -21,7 +22,9 @@ from mautrix.errors import (
 from mautrix.types import (
     JSON,
     BatchID,
+    BatchSendEvent,
     BatchSendResponse,
+    BatchSendStateEvent,
     ContentURI,
     EventContent,
     EventID,
@@ -30,7 +33,6 @@ from mautrix.types import (
     JoinRulesStateEventContent,
     Member,
     Membership,
-    MessageEvent,
     PowerLevelStateEventContent,
     PresenceState,
     RoomAvatarStateEventContent,
@@ -38,7 +40,7 @@ from mautrix.types import (
     RoomNameStateEventContent,
     RoomPinnedEventsStateEventContent,
     RoomTopicStateEventContent,
-    StateEvent,
+    SerializableAttrs,
     StateEventContent,
     UserID,
 )
@@ -91,6 +93,8 @@ ENSURE_JOINED_METHODS = (
 
 DOUBLE_PUPPET_SOURCE_KEY = "fi.mau.double_puppet_source"
 
+T = TypeVar("T")
+
 
 class IntentAPI(StoreUpdatingAPI):
     """
@@ -131,7 +135,9 @@ class IntentAPI(StoreUpdatingAPI):
                 room_id = kwargs.get("room_id", None)
                 if not room_id:
                     room_id = args[0]
-                await __self.ensure_joined(room_id)
+                ensure_joined = kwargs.pop("ensure_joined", True)
+                if ensure_joined:
+                    await __self.ensure_joined(room_id)
                 return await __method(*args, **kwargs)
 
             setattr(self, method.__name__, wrapper)
@@ -188,6 +194,13 @@ class IntentAPI(StoreUpdatingAPI):
     # endregion
     # region Room actions
 
+    def _add_source_key(self, content: T = None) -> T:
+        if self.api.is_real_user and self.api.bridge_name:
+            if not content:
+                content = {}
+            content[DOUBLE_PUPPET_SOURCE_KEY] = self.api.bridge_name
+        return content
+
     async def invite_user(
         self,
         room_id: RoomID,
@@ -226,37 +239,93 @@ class IntentAPI(StoreUpdatingAPI):
                 await self.state_store.get_membership(room_id, user_id) not in ok_states
             )
             if do_invite:
+                extra_content = self._add_source_key(extra_content)
                 await super().invite_user(
                     room_id, user_id, reason=reason, extra_content=extra_content
                 )
                 await self.state_store.invited(room_id, user_id)
+        except MAlreadyJoined as e:
+            await self.state_store.joined(room_id, user_id)
         except MatrixRequestError as e:
+            # TODO remove this once MSC3848 is released and minimum spec version is bumped
             if e.errcode == "M_FORBIDDEN" and "is already in the room" in e.message:
                 await self.state_store.joined(room_id, user_id)
             else:
-                raise IntentError(f"Failed to invite {user_id} to {room_id}", e)
+                raise
+
+    async def kick_user(
+        self,
+        room_id: RoomID,
+        user_id: UserID,
+        reason: str = "",
+        extra_content: dict[str, JSON] | None = None,
+    ) -> None:
+        extra_content = self._add_source_key(extra_content)
+        await super().kick_user(room_id, user_id, reason=reason, extra_content=extra_content)
+
+    async def ban_user(
+        self,
+        room_id: RoomID,
+        user_id: UserID,
+        reason: str = "",
+        extra_content: dict[str, JSON] | None = None,
+    ) -> None:
+        extra_content = self._add_source_key(extra_content)
+        await super().ban_user(room_id, user_id, reason=reason, extra_content=extra_content)
+
+    async def unban_user(
+        self,
+        room_id: RoomID,
+        user_id: UserID,
+        reason: str = "",
+        extra_content: dict[str, JSON] | None = None,
+    ) -> None:
+        extra_content = self._add_source_key(extra_content)
+        await super().unban_user(room_id, user_id, reason=reason, extra_content=extra_content)
+
+    async def join_room_by_id(
+        self,
+        room_id: RoomID,
+        third_party_signed: JSON = None,
+        extra_content: dict[str, JSON] | None = None,
+    ) -> RoomID:
+        extra_content = self._add_source_key(extra_content)
+        return await super().join_room_by_id(
+            room_id, third_party_signed=third_party_signed, extra_content=extra_content
+        )
+
+    async def leave_room(
+        self,
+        room_id: RoomID,
+        reason: str | None = None,
+        extra_content: dict[str, JSON] | None = None,
+        raise_not_in_room: bool = False,
+    ) -> None:
+        extra_content = self._add_source_key(extra_content)
+        await super().leave_room(room_id, reason, extra_content, raise_not_in_room)
 
     def set_room_avatar(
         self, room_id: RoomID, avatar_url: ContentURI | None, **kwargs
     ) -> Awaitable[EventID]:
-        return self.send_state_event(
-            room_id, EventType.ROOM_AVATAR, RoomAvatarStateEventContent(url=avatar_url), **kwargs
-        )
+        content = RoomAvatarStateEventContent(url=avatar_url)
+        content = self._add_source_key(content)
+        return self.send_state_event(room_id, EventType.ROOM_AVATAR, content, **kwargs)
 
     def set_room_name(self, room_id: RoomID, name: str, **kwargs) -> Awaitable[EventID]:
-        return self.send_state_event(
-            room_id, EventType.ROOM_NAME, RoomNameStateEventContent(name=name), **kwargs
-        )
+        content = RoomNameStateEventContent(name=name)
+        content = self._add_source_key(content)
+        return self.send_state_event(room_id, EventType.ROOM_NAME, content, **kwargs)
 
     def set_room_topic(self, room_id: RoomID, topic: str, **kwargs) -> Awaitable[EventID]:
-        return self.send_state_event(
-            room_id, EventType.ROOM_TOPIC, RoomTopicStateEventContent(topic=topic), **kwargs
-        )
+        content = RoomTopicStateEventContent(topic=topic)
+        content = self._add_source_key(content)
+        return self.send_state_event(room_id, EventType.ROOM_TOPIC, content, **kwargs)
 
     async def get_power_levels(
-        self, room_id: RoomID, ignore_cache: bool = False
+        self, room_id: RoomID, ignore_cache: bool = False, ensure_joined: bool = True
     ) -> PowerLevelStateEventContent:
-        await self.ensure_joined(room_id)
+        if ensure_joined:
+            await self.ensure_joined(room_id)
         if not ignore_cache:
             levels = await self.state_store.get_power_levels(room_id)
             if levels:
@@ -265,12 +334,17 @@ class IntentAPI(StoreUpdatingAPI):
             levels = await self.get_state_event(room_id, EventType.ROOM_POWER_LEVELS)
         except MNotFound:
             levels = PowerLevelStateEventContent()
+        except MForbidden:
+            if not ensure_joined:
+                return PowerLevelStateEventContent()
+            raise
         await self.state_store.set_power_levels(room_id, levels)
         return levels
 
     async def set_power_levels(
         self, room_id: RoomID, content: PowerLevelStateEventContent, **kwargs
     ) -> EventID:
+        content = self._add_source_key(content)
         response = await self.send_state_event(
             room_id, EventType.ROOM_POWER_LEVELS, content, **kwargs
         )
@@ -289,12 +363,9 @@ class IntentAPI(StoreUpdatingAPI):
     def set_pinned_messages(
         self, room_id: RoomID, events: list[EventID], **kwargs
     ) -> Awaitable[EventID]:
-        return self.send_state_event(
-            room_id,
-            EventType.ROOM_PINNED_EVENTS,
-            RoomPinnedEventsStateEventContent(pinned=events),
-            **kwargs,
-        )
+        content = RoomPinnedEventsStateEventContent(pinned=events)
+        content = self._add_source_key(content)
+        return self.send_state_event(room_id, EventType.ROOM_PINNED_EVENTS, content, **kwargs)
 
     async def pin_message(self, room_id: RoomID, event_id: EventID) -> None:
         events = await self.get_pinned_messages(room_id)
@@ -309,12 +380,9 @@ class IntentAPI(StoreUpdatingAPI):
             await self.set_pinned_messages(room_id, events)
 
     async def set_join_rule(self, room_id: RoomID, join_rule: JoinRule, **kwargs):
-        await self.send_state_event(
-            room_id,
-            EventType.ROOM_JOIN_RULES,
-            JoinRulesStateEventContent(join_rule=join_rule),
-            **kwargs,
-        )
+        content = JoinRulesStateEventContent(join_rule=join_rule)
+        content = self._add_source_key(content)
+        await self.send_state_event(room_id, EventType.ROOM_JOIN_RULES, content, **kwargs)
 
     async def get_room_displayname(
         self, room_id: RoomID, user_id: UserID, ignore_cache=False
@@ -358,10 +426,7 @@ class IntentAPI(StoreUpdatingAPI):
         self, room_id: RoomID, event_type: EventType, content: EventContent, **kwargs
     ) -> EventID:
         await self._ensure_has_power_level_for(room_id, event_type)
-
-        if self.api.is_real_user and self.api.bridge_name is not None:
-            content[DOUBLE_PUPPET_SOURCE_KEY] = self.api.bridge_name
-
+        content = self._add_source_key(content)
         return await super().send_message_event(room_id, event_type, content, **kwargs)
 
     async def redact(
@@ -373,10 +438,7 @@ class IntentAPI(StoreUpdatingAPI):
         **kwargs,
     ) -> EventID:
         await self._ensure_has_power_level_for(room_id, EventType.ROOM_REDACTION)
-        if self.api.is_real_user and self.api.bridge_name:
-            if not extra_content:
-                extra_content = {}
-            extra_content[DOUBLE_PUPPET_SOURCE_KEY] = self.api.bridge_name
+        extra_content = self._add_source_key(extra_content)
         return await super().redact(
             room_id, event_id, reason, extra_content=extra_content, **kwargs
         )
@@ -390,6 +452,7 @@ class IntentAPI(StoreUpdatingAPI):
         **kwargs,
     ) -> EventID:
         await self._ensure_has_power_level_for(room_id, event_type, state_key=state_key)
+        content = self._add_source_key(content)
         return await super().send_state_event(room_id, event_type, content, state_key, **kwargs)
 
     async def get_room_members(
@@ -435,8 +498,9 @@ class IntentAPI(StoreUpdatingAPI):
         prev_event_id: EventID,
         *,
         batch_id: BatchID | None = None,
-        events: Iterable[MessageEvent],
-        state_events_at_start: Iterable[StateEvent] = None,
+        events: Iterable[BatchSendEvent],
+        state_events_at_start: Iterable[BatchSendStateEvent] = (),
+        beeper_new_messages: bool = False,
     ) -> BatchSendResponse:
         """
         Send a batch of historical events into a room. See `MSC2716`_ for more info.
@@ -454,14 +518,18 @@ class IntentAPI(StoreUpdatingAPI):
             state_events_at_start: The state events to send at the start of the batch.
                                    These will be sent as outlier events, which means they won't be
                                    a part of the actual room state.
+            beeper_new_messages: Custom flag to tell the server that the messages can be sent to
+                                 the end of the room as normal messages instead of history.
 
         Returns:
             All the event IDs generated, plus a batch ID that can be passed back to this method.
         """
         path = Path.unstable["org.matrix.msc2716"].rooms[room_id].batch_send
-        query = {"prev_event_id": prev_event_id}
+        query: JSON = {"prev_event_id": prev_event_id}
         if batch_id:
             query["batch_id"] = batch_id
+        if beeper_new_messages:
+            query["com.beeper.new_messages"] = "true"
         resp = await self.api.request(
             Method.POST,
             path,
@@ -554,8 +622,6 @@ class IntentAPI(StoreUpdatingAPI):
             await self._register()
         except MUserInUse:
             pass
-        except MatrixRequestError as e:
-            raise IntentError(f"Failed to register {self.mxid}", e)
         await self.state_store.registered(self.mxid)
 
     async def _ensure_has_power_level_for(
@@ -571,7 +637,7 @@ class IntentAPI(StoreUpdatingAPI):
             return
         if not await self.state_store.has_power_levels_cached(room_id):
             # TODO add option to not try to fetch power levels from server
-            await self.get_power_levels(room_id, ignore_cache=True)
+            await self.get_power_levels(room_id, ignore_cache=True, ensure_joined=False)
         if not await self.state_store.has_power_level(room_id, self.mxid, event_type):
             # TODO implement something better
             raise IntentError(
