@@ -1,11 +1,11 @@
-# Copyright (c) 2021 Tulir Asokan
+# Copyright (c) 2022 Tulir Asokan
 #
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 from __future__ import annotations
 
-from typing import Any, Type
+from typing import Any
 from abc import ABC, abstractmethod
 import sys
 
@@ -17,7 +17,14 @@ from mautrix.appservice import AppService, ASStateStore
 from mautrix.client.state_store.asyncpg import PgStateStore as PgClientStateStore
 from mautrix.errors import MExclusive, MUnknownToken
 from mautrix.types import RoomID, UserID
-from mautrix.util.async_db import Database, UpgradeTable
+from mautrix.util.async_db import (
+    Database,
+    DatabaseException,
+    DatabaseNotOwned,
+    ForeignTablesFound,
+    UnsupportedDatabaseVersion,
+    UpgradeTable,
+)
 from mautrix.util.bridge_state import BridgeState, BridgeStateEvent, GlobalBridgeState
 from mautrix.util.program import Program
 
@@ -33,12 +40,12 @@ except ImportError:
 class Bridge(Program, ABC):
     db: Database
     az: AppService
-    state_store_class: Type[ASStateStore] = PgBridgeStateStore
+    state_store_class: type[ASStateStore] = PgBridgeStateStore
     state_store: ASStateStore
     upgrade_table: UpgradeTable
-    config_class: Type[br.BaseBridgeConfig]
+    config_class: type[br.BaseBridgeConfig]
     config: br.BaseBridgeConfig
-    matrix_class: Type[br.BaseMatrixHandler]
+    matrix_class: type[br.BaseMatrixHandler]
     matrix: br.BaseMatrixHandler
     repo_url: str
     markdown_version: str
@@ -51,9 +58,9 @@ class Bridge(Program, ABC):
         description: str = None,
         command: str = None,
         version: str = None,
-        config_class: Type[br.BaseBridgeConfig] = None,
-        matrix_class: Type[br.BaseMatrixHandler] = None,
-        state_store_class: Type[ASStateStore] = None,
+        config_class: type[br.BaseBridgeConfig] = None,
+        matrix_class: type[br.BaseMatrixHandler] = None,
+        state_store_class: type[ASStateStore] = None,
     ) -> None:
         super().__init__(module, name, description, command, version, config_class)
         if matrix_class:
@@ -81,6 +88,16 @@ class Bridge(Program, ABC):
                 "(not needed for running the bridge)"
             ),
         )
+        self.parser.add_argument(
+            "--ignore-unsupported-database",
+            action="store_true",
+            help="Run even if the database schema is too new",
+        )
+        self.parser.add_argument(
+            "--ignore-foreign-tables",
+            action="store_true",
+            help="Run even if the database contains tables from other programs (like Synapse)",
+        )
 
     def preinit(self) -> None:
         super().preinit()
@@ -119,7 +136,7 @@ class Bridge(Program, ABC):
 
     def prepare_appservice(self) -> None:
         self.make_state_store()
-        mb = 1024 ** 2
+        mb = 1024**2
         default_http_retry_count = self.config.get("homeserver.http_retry_count", None)
         if self.name not in HTTPAPI.default_ua:
             HTTPAPI.default_ua = f"{self.name}/{self.version} {HTTPAPI.default_ua}"
@@ -152,19 +169,34 @@ class Bridge(Program, ABC):
             self.config["appservice.database"],
             upgrade_table=self.upgrade_table,
             db_args=self.config["appservice.database_opts"],
+            owner_name=self.name,
+            ignore_foreign_tables=self.args.ignore_foreign_tables,
         )
 
     def prepare_bridge(self) -> None:
         self.matrix = self.matrix_class(bridge=self)
 
+    def _log_db_error(self, e: Exception) -> None:
+        self.log.critical("Failed to initialize database", exc_info=e)
+        if isinstance(e, DatabaseException) and e.explanation:
+            self.log.info(e.explanation)
+        sys.exit(25)
+
     async def start_db(self) -> None:
         if hasattr(self, "db") and isinstance(self.db, Database):
             self.log.debug("Starting database...")
-            await self.db.start()
-            if isinstance(self.state_store, PgClientStateStore):
-                await self.state_store.upgrade_table.upgrade(self.db)
-            if self.matrix.e2ee:
-                self.matrix.e2ee.crypto_db.override_pool(self.db)
+            ignore_unsupported = self.args.ignore_unsupported_database
+            self.db.upgrade_table.allow_unsupported = ignore_unsupported
+            try:
+                await self.db.start()
+                if isinstance(self.state_store, PgClientStateStore):
+                    self.state_store.upgrade_table.allow_unsupported = ignore_unsupported
+                    await self.state_store.upgrade_table.upgrade(self.db)
+                if self.matrix.e2ee:
+                    self.matrix.e2ee.crypto_db.allow_unsupported = ignore_unsupported
+                    self.matrix.e2ee.crypto_db.override_pool(self.db)
+            except Exception as e:
+                self._log_db_error(e)
 
     async def stop_db(self) -> None:
         if hasattr(self, "db") and isinstance(self.db, Database):
@@ -201,6 +233,11 @@ class Bridge(Program, ABC):
             state = BridgeState(state_event=BridgeStateEvent.UNCONFIGURED).fill()
             await state.send(status_endpoint, self.az.as_token, self.log)
 
+    async def system_exit(self) -> None:
+        if hasattr(self, "db") and isinstance(self.db, Database):
+            self.log.trace("Stopping database due to SystemExit")
+            await self.db.stop()
+
     async def stop(self) -> None:
         if self.manhole:
             self.manhole.close()
