@@ -11,7 +11,7 @@ import sys
 
 from mautrix import __optional_imports__
 from mautrix.appservice import AppService
-from mautrix.client import Client, SyncStore
+from mautrix.client import Client, InternalEventType, SyncStore
 from mautrix.crypto import CryptoStore, OlmMachine, PgCryptoStore, RejectKeyShare, StateStore
 from mautrix.errors import EncryptionError, SessionNotFound
 from mautrix.types import (
@@ -58,6 +58,9 @@ class EncryptionManager:
     crypto_db: Database | None
     state_store: StateStore
 
+    min_send_trust: TrustState
+    key_sharing_enabled: bool
+
     bridge: br.Bridge
     az: AppService
     _id_prefix: str
@@ -72,7 +75,6 @@ class EncryptionManager:
         user_id_prefix: str,
         user_id_suffix: str,
         db_url: str,
-        key_sharing_config: dict[str, bool] = None,
     ) -> None:
         self.loop = bridge.loop or asyncio.get_event_loop()
         self.bridge = bridge
@@ -81,7 +83,6 @@ class EncryptionManager:
         self._id_prefix = user_id_prefix
         self._id_suffix = user_id_suffix
         self._share_session_events = {}
-        self.key_sharing_config = key_sharing_config or {}
         pickle_key = "mautrix.bridge.e2ee"
         self.crypto_db = Database.create(
             url=db_url,
@@ -100,12 +101,21 @@ class EncryptionManager:
             default_retry_count=default_http_retry_count,
         )
         self.crypto = OlmMachine(self.client, self.crypto_store, self.state_store)
+        self.client.add_event_handler(InternalEventType.SYNC_STOPPED, self._exit_on_sync_fail)
         self.crypto.allow_key_share = self.allow_key_share
+        verification_levels = bridge.config["bridge.encryption.verification_levels"]
+        self.min_send_trust = TrustState.parse(verification_levels["send"])
+        self.crypto.share_keys_min_trust = TrustState.parse(verification_levels["share"])
+        self.crypto.send_keys_min_trust = TrustState.parse(verification_levels["receive"])
+        self.key_sharing_enabled = bridge.config["bridge.encryption.allow_key_sharing"]
+
+    async def _exit_on_sync_fail(self, data) -> None:
+        if data["error"]:
+            self.log.critical("Exiting due to crypto sync error")
+            sys.exit(32)
 
     async def allow_key_share(self, device: DeviceIdentity, request: RequestedKeyInfo) -> bool:
-        require_verification = self.key_sharing_config.get("require_verification", True)
-        allow = self.key_sharing_config.get("allow", False)
-        if not allow:
+        if not self.key_sharing_enabled:
             self.log.debug(
                 f"Key sharing not enabled, ignoring key request from "
                 f"{device.user_id}/{device.device_id}"
@@ -118,7 +128,7 @@ class EncryptionManager:
                 code=RoomKeyWithheldCode.BLACKLISTED,
                 reason="You have been blacklisted by this device",
             )
-        elif device.trust == TrustState.VERIFIED or not require_verification:
+        elif device.trust >= self.crypto.share_keys_min_trust:
             portal = await self.bridge.get_portal(request.room_id)
             if portal is None:
                 raise RejectKeyShare(
@@ -200,7 +210,7 @@ class EncryptionManager:
                 f" waiting {wait_session_timeout} seconds..."
             )
             got_keys = await self.crypto.wait_for_session(
-                evt.room_id, e.sender_key, e.session_id, timeout=wait_session_timeout
+                evt.room_id, e.session_id, timeout=wait_session_timeout
             )
             if got_keys:
                 self.log.debug(
@@ -244,9 +254,31 @@ class EncryptionManager:
         if not device_id:
             await self.crypto_store.put_device_id(self.client.device_id)
             self.log.debug(f"Logged in with new device ID {self.client.device_id}")
+        elif self.crypto.account.shared:
+            await self._verify_keys_are_on_server()
         _ = self.client.start(self._filter)
         self.log.info("End-to-bridge encryption support is enabled")
 
+    async def _verify_keys_are_on_server(self) -> None:
+        self.log.debug("Making sure keys are still on server")
+        try:
+            resp = await self.client.query_keys([self.client.mxid])
+        except Exception:
+            self.log.critical(
+                "Failed to query own keys to make sure device still exists", exc_info=True
+            )
+            sys.exit(33)
+        try:
+            own_keys = resp.device_keys[self.client.mxid][self.client.device_id]
+            if len(own_keys.keys) > 0:
+                return
+        except KeyError:
+            pass
+        self.log.critical("Existing device doesn't have keys on server, resetting crypto")
+        await self.crypto.crypto_store.delete()
+        await self.client.logout_all()
+        sys.exit(34)
+
     async def stop(self) -> None:
         self.client.stop()
         await self.crypto_store.close()
