@@ -23,6 +23,7 @@ from mautrix.errors import (
 )
 from mautrix.types import (
     BaseRoomEvent,
+    BeeperMessageStatusEventContent,
     EncryptedEvent,
     Event,
     EventID,
@@ -32,20 +33,27 @@ from mautrix.types import (
     MemberStateEventContent,
     MessageEvent,
     MessageEventContent,
+    MessageStatus,
+    MessageStatusReason,
     MessageType,
     PresenceEvent,
     ReactionEvent,
     ReceiptEvent,
     ReceiptType,
     RedactionEvent,
+    RelatesTo,
+    RelationType,
     RoomID,
     RoomType,
     SingleReceiptEventContent,
+    SpecVersions,
     StateEvent,
     StateUnsigned,
     TextMessageEventContent,
+    TrustState,
     TypingEvent,
     UserID,
+    Version,
     VersionsResponse,
 )
 from mautrix.util import markdown
@@ -86,6 +94,44 @@ EVENT_TIME = Histogram(
 )
 
 
+class UnencryptedMessageError(DecryptionError):
+    def __init__(self) -> None:
+        super().__init__("unencrypted message")
+
+    @property
+    def human_message(self) -> str:
+        return "the message is not encrypted"
+
+
+class EncryptionUnsupportedError(DecryptionError):
+    def __init__(self) -> None:
+        super().__init__("encryption is not supported")
+
+    @property
+    def human_message(self) -> str:
+        return "the bridge is not configured to support encryption"
+
+
+class DeviceUntrustedError(DecryptionError):
+    def __init__(self, trust: TrustState) -> None:
+        explanation = {
+            TrustState.BLACKLISTED: "device is blacklisted",
+            TrustState.UNVERIFIED: "unverified",
+            TrustState.UNKNOWN_DEVICE: "device info not found",
+            TrustState.FORWARDED: "keys were forwarded from an unknown device",
+            TrustState.CROSS_SIGNED_UNTRUSTED: (
+                "cross-signing keys changed after setting up the bridge"
+            ),
+        }.get(trust)
+        base = "your device is not trusted"
+        self.message = f"{base} ({explanation})" if explanation else base
+        super().__init__(self.message)
+
+    @property
+    def human_message(self) -> str:
+        return self.message
+
+
 class BaseMatrixHandler:
     log: TraceLogger = logging.getLogger("mau.mx")
     az: AppService
@@ -93,8 +139,10 @@ class BaseMatrixHandler:
     config: config.BaseBridgeConfig
     bridge: br.Bridge
     e2ee: EncryptionManager | None
+    require_e2ee: bool
     media_config: MediaRepoConfig
     versions: VersionsResponse
+    minimum_spec_version: Version = SpecVersions.V11
 
     user_id_prefix: str
     user_id_suffix: str
@@ -109,10 +157,11 @@ class BaseMatrixHandler:
         self.bridge = bridge
         self.commands = command_processor or cmd.CommandProcessor(bridge=bridge)
         self.media_config = MediaRepoConfig(upload_size=50 * 1024 * 1024)
-        self.versions = VersionsResponse(versions=["v1.2"])
+        self.versions = VersionsResponse.deserialize({"versions": ["v1.3"]})
         self.az.matrix_event_handler(self.int_handle_event)
 
         self.e2ee = None
+        self.require_e2ee = False
         if self.config["bridge.encryption.allow"]:
             if not EncryptionManager:
                 self.log.fatal(
@@ -133,8 +182,8 @@ class BaseMatrixHandler:
                 user_id_suffix=self.user_id_suffix,
                 homeserver_address=self.config["homeserver.address"],
                 db_url=self.config["appservice.database"],
-                key_sharing_config=self.config["bridge.encryption.key_sharing"],
             )
+            self.require_e2ee = self.config["bridge.encryption.require"]
 
         self.management_room_text = self.config.get(
             "bridge.management_room_text",
@@ -149,6 +198,16 @@ class BaseMatrixHandler:
             False,
         )
 
+    async def check_versions(self) -> None:
+        if not self.versions.supports_at_least(self.minimum_spec_version):
+            self.log.fatal(
+                "The homeserver is outdated "
+                "(server supports Matrix %s, but the bridge requires at least %s)",
+                self.versions.latest_version,
+                self.minimum_spec_version,
+            )
+            sys.exit(18)
+
     async def wait_for_connection(self) -> None:
         self.log.info("Ensuring connectivity to homeserver")
         errors = 0
@@ -156,6 +215,7 @@ class BaseMatrixHandler:
         while True:
             try:
                 self.versions = await self.az.intent.versions()
+                await self.check_versions()
                 await self.az.intent.whoami()
                 break
             except (MUnknownToken, MExclusive):
@@ -243,6 +303,26 @@ class BaseMatrixHandler:
     async def handle_join(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
         pass
 
+    async def handle_knock(
+        self, room_id: RoomID, user_id: UserID, reason: str, event_id: EventID
+    ) -> None:
+        pass
+
+    async def handle_retract_knock(
+        self, room_id: RoomID, user_id: UserID, reason: str, event_id: EventID
+    ) -> None:
+        pass
+
+    async def handle_reject_knock(
+        self, room_id: RoomID, user_id: UserID, sender: UserID, reason: str, event_id: EventID
+    ) -> None:
+        pass
+
+    async def handle_accept_knock(
+        self, room_id: RoomID, user_id: UserID, sender: UserID, reason: str, event_id: EventID
+    ) -> None:
+        pass
+
     async def handle_member_info_change(
         self,
         room_id: RoomID,
@@ -298,7 +378,7 @@ class BaseMatrixHandler:
             return
         if create_evt.type == RoomType.SPACE:
             await self.handle_puppet_space_invite(room_id, puppet, invited_by, evt)
-        elif len(members) > 2:
+        elif len(members) > 2 or not evt.content.is_direct:
             await self.handle_puppet_group_invite(room_id, puppet, invited_by, evt, members)
         else:
             await self.handle_puppet_dm_invite(room_id, puppet, invited_by, evt)
@@ -461,33 +541,79 @@ class BaseMatrixHandler:
             text = text[len(prefix) + 1 :].lstrip()
         return is_command, text
 
-    async def handle_message(
-        self, room_id: RoomID, user_id: UserID, message: MessageEventContent, event_id: EventID
-    ) -> None:
-        async def bail(error_text: str, step=MessageSendCheckpointStep.REMOTE) -> None:
-            self.log.debug(error_text)
-            await MessageSendCheckpoint(
-                event_id=event_id,
-                room_id=room_id,
-                step=step,
-                timestamp=int(time.time() * 1000),
-                status=MessageSendCheckpointStatus.PERM_FAILURE,
-                reported_by=MessageSendCheckpointReportedBy.BRIDGE,
-                event_type=EventType.ROOM_MESSAGE,
-                message_type=message.msgtype,
-                info=error_text,
-            ).send(
-                self.bridge.config["homeserver.message_send_checkpoint_endpoint"],
-                self.az.as_token,
-                self.log,
+    async def _send_crypto_status_error(
+        self,
+        evt: Event,
+        err: DecryptionError | None = None,
+        retry_num: int = 0,
+        is_final: bool = True,
+        edit: EventID | None = None,
+        wait_for: int | None = None,
+    ) -> EventID | None:
+        msg = str(err)
+        if isinstance(err, (SessionNotFound, UnencryptedMessageError)):
+            msg = err.human_message
+        self._send_message_checkpoint(
+            evt, MessageSendCheckpointStep.DECRYPTED, msg, permanent=is_final, retry_num=retry_num
+        )
+
+        if wait_for:
+            msg += f". The bridge will retry for {wait_for} seconds"
+        full_msg = f"\u26a0 Your message was not bridged: {msg}."
+        if isinstance(err, EncryptionUnsupportedError):
+            full_msg = "🔒️ This bridge has not been configured to support encryption"
+        event_id = None
+        if self.config.get("bridge.delivery_error_reports", True):
+            try:
+                content = TextMessageEventContent(msgtype=MessageType.NOTICE, body=full_msg)
+                if edit:
+                    content.set_edit(edit)
+                event_id = await self.az.intent.send_message(evt.room_id, content)
+            except IntentError:
+                self.log.debug("IntentError while sending encryption error", exc_info=True)
+                self.log.error(
+                    "Got IntentError while trying to send encryption error message. "
+                    "This likely means the bridge bot is not in the room, which can "
+                    "happen if you force-enable e2ee on the homeserver without enabling "
+                    "it by default on the bridge (bridge -> encryption -> default)."
+                )
+
+        if self.config.get("bridge.message_status_events", False):
+            status_content = BeeperMessageStatusEventContent(
+                network="",  # TODO set network properly
+                relates_to=RelatesTo(rel_type=RelationType.REFERENCE, event_id=evt.event_id),
+                status=MessageStatus.RETRIABLE if is_final else MessageStatus.PENDING,
+                reason=MessageStatusReason.UNDECRYPTABLE,
+                error=msg,
+                message=err.human_message if err else None,
             )
+            status_content.fill_legacy_booleans()
+            await self.az.intent.send_message_event(
+                evt.room_id, EventType.BEEPER_MESSAGE_STATUS, status_content
+            )
+
+        return event_id
+
+    async def handle_message(self, evt: MessageEvent, was_encrypted: bool = False) -> None:
+        room_id = evt.room_id
+        user_id = evt.sender
+        event_id = evt.event_id
+        message = evt.content
+
+        if not was_encrypted and self.require_e2ee:
+            self.log.warning(f"Dropping {event_id} from {user_id} as it's not encrypted!")
+            await self._send_crypto_status_error(evt, UnencryptedMessageError(), 0)
+            return
 
         sender = await self.bridge.get_user(user_id)
         if not sender or not await self.allow_message(sender):
-            await bail(
+            self.log.debug(
                 f"Ignoring message {event_id} from {user_id} to {room_id}:"
                 " user is not whitelisted."
             )
+            self._send_message_checkpoint(
+                evt, MessageSendCheckpointStep.BRIDGE, "user is not whitelisted"
+            )
             return
         self.log.debug(f"Received Matrix event {event_id} from {sender.mxid} in {room_id}")
         self.log.trace("Event %s content: %s", event_id, message)
@@ -501,19 +627,31 @@ class BaseMatrixHandler:
             if await self.allow_bridging_message(sender, portal):
                 await portal.handle_matrix_message(sender, message, event_id)
             else:
-                await bail(
+                self.log.debug(
                     f"Ignoring event {event_id} from {sender.mxid}:"
                     " not allowed to send to portal"
                 )
+                self._send_message_checkpoint(
+                    evt,
+                    MessageSendCheckpointStep.BRIDGE,
+                    "user is not allowed to send to the portal",
+                )
             return
 
         if message.msgtype != MessageType.TEXT:
-            await bail(f"Ignoring event {event_id}: not a portal room and not a m.text message")
+            self.log.debug(
+                f"Ignoring event {event_id}: not a portal room and not a m.text message"
+            )
+            self._send_message_checkpoint(
+                evt, MessageSendCheckpointStep.BRIDGE, "not a portal room and not a m.text message"
+            )
             return
         elif not await self.allow_command(sender):
-            await bail(
-                f"Ignoring command {event_id} from {sender.mxid}: not allowed to perform command",
-                step=MessageSendCheckpointStep.COMMAND,
+            self.log.debug(
+                f"Ignoring command {event_id} from {sender.mxid}: not allowed to run commands"
+            )
+            self._send_message_checkpoint(
+                evt, MessageSendCheckpointStep.COMMAND, "not allowed to run commands"
             )
             return
 
@@ -542,7 +680,8 @@ class BaseMatrixHandler:
                     bridge_bot_in_room,
                 )
             except Exception as e:
-                await bail(repr(e), step=MessageSendCheckpointStep.COMMAND)
+                self.log.debug(f"Error handling command {command} from {sender}: {e}")
+                self._send_message_checkpoint(evt, MessageSendCheckpointStep.COMMAND, e)
             else:
                 await MessageSendCheckpoint(
                     event_id=event_id,
@@ -559,10 +698,13 @@ class BaseMatrixHandler:
                     self.log,
                 )
         else:
-            await bail(
+            self.log.debug(
                 f"Ignoring event {event_id} from {sender.mxid}:"
                 " not a command and not a portal room"
             )
+            self._send_message_checkpoint(
+                evt, MessageSendCheckpointStep.COMMAND, "not a command and not a portal room"
+            )
 
     async def _is_direct_chat(self, room_id: RoomID) -> tuple[bool, bool]:
         try:
@@ -573,7 +715,7 @@ class BaseMatrixHandler:
 
     async def handle_receipt(self, evt: ReceiptEvent) -> None:
         for event_id, receipts in evt.content.items():
-            for user_id, data in receipts[ReceiptType.READ].items():
+            for user_id, data in receipts.get(ReceiptType.READ, {}).items():
                 user = await self.bridge.get_user(user_id, create=False)
                 if not user or not await user.is_logged_in():
                     continue
@@ -610,41 +752,49 @@ class BaseMatrixHandler:
         except Exception:
             self.log.exception("Error handling manually received Matrix event")
 
-    async def send_encryption_error_notice(
-        self, evt: EncryptedEvent, error: DecryptionError
+    async def _post_decrypt(
+        self, evt: Event, retry_num: int = 0, error_event_id: EventID | None = None
     ) -> None:
-        await self.az.intent.send_notice(
-            evt.room_id, f"\u26a0 Your message was not bridged: {error}"
+        trust_state = evt["mautrix"]["trust_state"]
+        if trust_state < self.e2ee.min_send_trust:
+            self.log.warning(
+                f"Dropping {evt.event_id} from {evt.sender} due to insufficient verification level"
+                f" (event: {trust_state}, required: {self.e2ee.min_send_trust})"
+            )
+            await self._send_crypto_status_error(
+                evt,
+                retry_num=retry_num,
+                err=DeviceUntrustedError(trust_state),
+                edit=error_event_id,
+            )
+            return
+
+        self._send_message_checkpoint(
+            evt, MessageSendCheckpointStep.DECRYPTED, retry_num=retry_num
         )
+        if error_event_id:
+            await self.az.intent.redact(evt.room_id, error_event_id)
+        await self.int_handle_event(evt, was_encrypted=True)
 
     async def handle_encrypted(self, evt: EncryptedEvent) -> None:
         if not self.e2ee:
-            self.send_decrypted_checkpoint(evt, "Encryption unsupported", True)
-            await self.handle_encrypted_unsupported(evt)
+            self.log.debug(
+                "Got encrypted message %s from %s, but encryption is not enabled",
+                evt.event_id,
+                evt.sender,
+            )
+            await self._send_crypto_status_error(evt, EncryptionUnsupportedError())
             return
         try:
-            decrypted = await self.e2ee.decrypt(evt, wait_session_timeout=5)
+            decrypted = await self.e2ee.decrypt(evt, wait_session_timeout=3)
         except SessionNotFound as e:
-            self.send_decrypted_checkpoint(evt, e, False)
-            await self._handle_encrypted_wait(evt, e, wait=15)
+            await self._handle_encrypted_wait(evt, e, wait=6)
         except DecryptionError as e:
             self.log.warning(f"Failed to decrypt {evt.event_id}: {e}")
             self.log.trace("%s decryption traceback:", evt.event_id, exc_info=True)
-            self.send_decrypted_checkpoint(evt, e, True)
-            await self.send_encryption_error_notice(evt, e)
+            await self._send_crypto_status_error(evt, e)
         else:
-            self.send_decrypted_checkpoint(decrypted)
-            await self.int_handle_event(decrypted, send_bridge_checkpoint=False)
-
-    async def handle_encrypted_unsupported(self, evt: EncryptedEvent) -> None:
-        self.log.debug(
-            "Got encrypted message %s from %s, but encryption is not enabled",
-            evt.event_id,
-            evt.sender,
-        )
-        await self.az.intent.send_notice(
-            evt.room_id, "🔒️ This bridge has not been configured to support encryption"
-        )
+            await self._post_decrypt(decrypted)
 
     async def _handle_encrypted_wait(
         self, evt: EncryptedEvent, err: SessionNotFound, wait: int
@@ -653,11 +803,6 @@ class BaseMatrixHandler:
             f"Couldn't find session {err.session_id} trying to decrypt {evt.event_id},"
             " waiting even longer"
         )
-        msg = (
-            "\u26a0 Your message was not bridged: the bridge hasn't received the decryption "
-            f"keys. The bridge will retry for {wait} seconds. If this error keeps happening, "
-            "try restarting your client."
-        )
         asyncio.create_task(
             self.e2ee.crypto.request_room_key(
                 evt.room_id,
@@ -666,19 +811,9 @@ class BaseMatrixHandler:
                 from_devices={evt.sender: [evt.content.device_id]},
             )
         )
-        try:
-            event_id = await self.az.intent.send_notice(evt.room_id, msg)
-        except IntentError:
-            self.log.debug("IntentError while sending encryption error", exc_info=True)
-            self.log.error(
-                "Got IntentError while trying to send encryption error message. "
-                "This likely means the bridge bot is not in the room, which can "
-                "happen if you force-enable e2ee on the homeserver without enabling "
-                "it by default on the bridge (bridge -> encryption -> default)."
-            )
-            return
+        event_id = await self._send_crypto_status_error(evt, err, is_final=False, wait_for=wait)
         got_keys = await self.e2ee.crypto.wait_for_session(
-            evt.room_id, err.sender_key, err.session_id, timeout=wait
+            evt.room_id, err.session_id, timeout=wait
         )
         if got_keys:
             self.log.debug(
@@ -688,26 +823,17 @@ class BaseMatrixHandler:
             try:
                 decrypted = await self.e2ee.decrypt(evt, wait_session_timeout=0)
             except DecryptionError as e:
-                self.send_decrypted_checkpoint(evt, e, True, retry_num=1)
+                await self._send_crypto_status_error(evt, e, retry_num=1, edit=event_id)
                 self.log.warning(f"Failed to decrypt {evt.event_id}: {e}")
                 self.log.trace("%s decryption traceback:", evt.event_id, exc_info=True)
-                msg = f"\u26a0 Your message was not bridged: {e}"
             else:
-                self.send_decrypted_checkpoint(decrypted, retry_num=1)
-                await self.az.intent.redact(evt.room_id, event_id)
-                await self.int_handle_event(decrypted, send_bridge_checkpoint=False)
+                await self._post_decrypt(decrypted, retry_num=1, error_event_id=event_id)
                 return
         else:
-            error_message = f"Didn't get {err.session_id}, giving up on {evt.event_id}"
-            self.log.warning(error_message)
-            self.send_decrypted_checkpoint(evt, error_message, True, retry_num=1)
-            msg = (
-                "\u26a0 Your message was not bridged: the bridge hasn't received the decryption"
-                " keys. If this error keeps happening, try restarting your client."
+            self.log.warning(f"Didn't get {err.session_id}, giving up on {evt.event_id}")
+            await self._send_crypto_status_error(
+                evt, SessionNotFound(err.session_id), retry_num=1, edit=event_id
             )
-        content = TextMessageEventContent(msgtype=MessageType.NOTICE, body=msg)
-        content.set_edit(event_id)
-        await self.az.intent.send_message(evt.room_id, content)
 
     async def handle_encryption(self, evt: StateEvent) -> None:
         await self.az.state_store.set_encryption_info(evt.room_id, evt.content)
@@ -719,12 +845,12 @@ class BaseMatrixHandler:
                 portal.log.debug("Received encryption event in direct portal: %s", evt.content)
                 await portal.enable_dm_encryption()
 
-    def send_message_send_checkpoint(
+    def _send_message_checkpoint(
         self,
         evt: Event,
         step: MessageSendCheckpointStep,
         err: Exception | str | None = None,
-        permanent: bool = False,
+        permanent: bool = True,
         retry_num: int = 0,
     ) -> None:
         endpoint = self.bridge.config["homeserver.message_send_checkpoint_endpoint"]
@@ -756,20 +882,6 @@ class BaseMatrixHandler:
         )
         asyncio.create_task(checkpoint.send(endpoint, self.az.as_token, self.log))
 
-    def send_bridge_checkpoint(self, evt: Event) -> None:
-        self.send_message_send_checkpoint(evt, MessageSendCheckpointStep.BRIDGE)
-
-    def send_decrypted_checkpoint(
-        self,
-        evt: Event,
-        err: Exception | str | None = None,
-        permanent: bool = False,
-        retry_num: int = 0,
-    ) -> None:
-        self.send_message_send_checkpoint(
-            evt, MessageSendCheckpointStep.DECRYPTED, err, permanent, retry_num
-        )
-
     allowed_event_classes: tuple[type, ...] = (
         MessageEvent,
         StateEvent,
@@ -800,15 +912,15 @@ class BaseMatrixHandler:
         # For non-room events and non-bridge-originated room events, allow.
         return True
 
-    async def int_handle_event(self, evt: Event, send_bridge_checkpoint: bool = True) -> None:
+    async def int_handle_event(self, evt: Event, was_encrypted: bool = False) -> None:
         if isinstance(evt, StateEvent) and evt.type == EventType.ROOM_MEMBER and self.e2ee:
             await self.e2ee.handle_member_event(evt)
         if not await self.allow_matrix_event(evt):
             return
         self.log.trace("Received event: %s", evt)
 
-        if send_bridge_checkpoint:
-            self.send_bridge_checkpoint(evt)
+        if not was_encrypted:
+            self._send_message_checkpoint(evt, MessageSendCheckpointStep.BRIDGE)
         start_time = time.time()
 
         if evt.type == EventType.ROOM_MEMBER:
@@ -817,7 +929,16 @@ class BaseMatrixHandler:
             prev_content = unsigned.prev_content or MemberStateEventContent()
             prev_membership = prev_content.membership if prev_content else Membership.JOIN
             if evt.content.membership == Membership.INVITE:
-                await self.int_handle_invite(evt)
+                if prev_membership == Membership.KNOCK:
+                    await self.handle_accept_knock(
+                        evt.room_id,
+                        UserID(evt.state_key),
+                        evt.sender,
+                        evt.content.reason,
+                        evt.event_id,
+                    )
+                else:
+                    await self.int_handle_invite(evt)
             elif evt.content.membership == Membership.LEAVE:
                 if prev_membership == Membership.BAN:
                     await self.handle_unban(
@@ -840,6 +961,20 @@ class BaseMatrixHandler:
                             evt.content.reason,
                             evt.event_id,
                         )
+                elif prev_membership == Membership.KNOCK:
+                    if evt.sender == evt.state_key:
+                        await self.handle_retract_knock(
+                            evt.room_id, UserID(evt.state_key), evt.content.reason, evt.event_id
+                        )
+                    else:
+                        await self.handle_reject_knock(
+                            evt.room_id,
+                            UserID(evt.state_key),
+                            evt.sender,
+                            evt.content.reason,
+                            evt.event_id,
+                        )
+
                 elif evt.sender == evt.state_key:
                     await self.handle_leave(evt.room_id, UserID(evt.state_key), evt.event_id)
                 else:
@@ -865,11 +1000,18 @@ class BaseMatrixHandler:
                     await self.handle_member_info_change(
                         evt.room_id, UserID(evt.state_key), evt.content, prev_content, evt.event_id
                     )
+            elif evt.content.membership == Membership.KNOCK:
+                await self.handle_knock(
+                    evt.room_id,
+                    UserID(evt.state_key),
+                    evt.content.reason,
+                    evt.event_id,
+                )
         elif evt.type in (EventType.ROOM_MESSAGE, EventType.STICKER):
             evt: MessageEvent
             if evt.type != EventType.ROOM_MESSAGE:
                 evt.content.msgtype = MessageType(str(evt.type))
-            await self.handle_message(evt.room_id, evt.sender, evt.content, evt.event_id)
+            await self.handle_message(evt, was_encrypted=was_encrypted)
         elif evt.type == EventType.ROOM_ENCRYPTED:
             await self.handle_encrypted(evt)
         elif evt.type == EventType.ROOM_ENCRYPTION:
