@@ -11,7 +11,13 @@ import asyncio
 from multidict import CIMultiDict
 
 from mautrix.api import Method, Path
-from mautrix.errors import MatrixRequestError, MatrixResponseError, MNotFound, MRoomInUse
+from mautrix.errors import (
+    MatrixRequestError,
+    MatrixResponseError,
+    MNotFound,
+    MNotJoined,
+    MRoomInUse,
+)
 from mautrix.types import (
     JSON,
     DirectoryPaginationToken,
@@ -391,7 +397,7 @@ class RoomMethods(EventMethods, BaseClientAPI):
             content[key] = value
         content = await self.fill_member_event(room_id, user_id, content) or content
         return await self.send_state_event(
-            room_id, EventType.ROOM_MEMBER, content=content, state_key=user_id
+            room_id, EventType.ROOM_MEMBER, content=content, state_key=user_id, ensure_joined=False
         )
 
     async def invite_user(
@@ -478,10 +484,53 @@ class RoomMethods(EventMethods, BaseClientAPI):
                 if reason:
                     data["reason"] = reason
                 await self.api.request(Method.POST, Path.v3.rooms[room_id].leave, content=data)
+        except MNotJoined:
+            if raise_not_in_room:
+                raise
         except MatrixRequestError as e:
+            # TODO remove this once MSC3848 is released and minimum spec version is bumped
             if "not in room" not in e.message or raise_not_in_room:
                 raise
 
+    async def knock_room(
+        self,
+        room_id_or_alias: RoomID | RoomAlias,
+        reason: str | None = None,
+        servers: list[str] | None = None,
+    ) -> RoomID:
+        """
+        Knock on a room, i.e. request to join it by its ID or alias, with an optional list of
+        servers to ask about the ID from.
+
+        See also: `API reference <https://spec.matrix.org/v1.3/client-server-api/#post_matrixclientv3knockroomidoralias>`__
+
+        Args:
+            room_id_or_alias: The ID of the room to knock on, or an alias pointing to the room.
+            reason: The reason for knocking on the room. This will be supplied as the ``reason`` on
+                the updated `m.room.member`_ event.
+            servers: A list of servers to ask about the room ID to knock. Not applicable for aliases,
+                as aliases already contain the necessary server information.
+
+        Returns:
+            The ID of the room the user knocked on.
+        """
+        data = {}
+        if reason:
+            data["reason"] = reason
+        query_params = CIMultiDict()
+        for server_name in servers or []:
+            query_params.add("server_name", server_name)
+        content = await self.api.request(
+            Method.POST,
+            Path.v3.knock[room_id_or_alias],
+            content=data,
+            query_params=query_params,
+        )
+        try:
+            return content["room_id"]
+        except KeyError:
+            raise MatrixResponseError("`room_id` not in response.")
+
     async def forget_room(self, room_id: RoomID) -> None:
         """
         Stop remembering a particular room, i.e. forget it.
@@ -581,7 +630,13 @@ class RoomMethods(EventMethods, BaseClientAPI):
             Method.POST, Path.v3.rooms[room_id].ban, {"user_id": user_id, "reason": reason}
         )
 
-    async def unban_user(self, room_id: RoomID, user_id: UserID) -> None:
+    async def unban_user(
+        self,
+        room_id: RoomID,
+        user_id: UserID,
+        reason: str = "",
+        extra_content: dict[str, JSON] | None = None,
+    ) -> None:
         """
         Unban a user from the room. This allows them to be invited to the room, and join if they
         would otherwise be allowed to join according to its join rules. The caller must have the
@@ -592,8 +647,22 @@ class RoomMethods(EventMethods, BaseClientAPI):
         Args:
             room_id: The ID of the room from which the user should be unbanned.
             user_id: The fully qualified user ID of the user being banned.
+            reason: The reason the user has been unbanned. This will be supplied as the ``reason`` on
+                the target's updated `m.room.member`_ event.
+            extra_content: Additional properties for the unban (leave) event content.
+                If a non-empty dict is passed, the unban will be created using
+                the ``PUT /state/m.room.member/...`` endpoint instead of ``POST /unban``.
         """
-        await self.api.request(Method.POST, Path.v3.rooms[room_id].unban, {"user_id": user_id})
+        if extra_content:
+            if reason and "reason" not in extra_content:
+                extra_content["reason"] = reason
+            await self.send_member_event(
+                room_id, user_id, Membership.LEAVE, extra_content=extra_content
+            )
+            return
+        await self.api.request(
+            Method.POST, Path.v3.rooms[room_id].unban, {"user_id": user_id, "reason": reason}
+        )
 
     # endregion
 
