@@ -5,7 +5,7 @@
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 from __future__ import annotations
 
-from typing import Any, Awaitable, Callable, Dict, TypedDict
+from typing import Any, Awaitable, Callable, TypedDict
 import asyncio
 import functools
 import json
@@ -16,10 +16,12 @@ from mautrix.types import (
     DeviceID,
     EncryptionKeyAlgorithm,
     IdentityKey,
+    KeyID,
     RequestedKeyInfo,
     RoomID,
     SessionID,
     SigningKey,
+    TrustState,
     UserID,
 )
 from mautrix.util.logging import TraceLogger
@@ -28,7 +30,7 @@ from .. import client as cli, crypto
 
 
 class SignedObject(TypedDict):
-    signatures: Dict[UserID, Dict[str, str]]
+    signatures: dict[UserID, dict[str, str]]
     unsigned: Any
 
 
@@ -40,19 +42,21 @@ class BaseOlmMachine:
 
     account: account.OlmAccount
 
-    allow_unverified_devices: bool
-    share_to_unverified_devices: bool
+    send_keys_min_trust: TrustState
+    share_keys_min_trust: TrustState
     allow_key_share: Callable[[crypto.DeviceIdentity, RequestedKeyInfo], Awaitable[bool]]
 
     # Futures that wait for responses to a key request
-    _key_request_waiters: Dict[SessionID, asyncio.Future]
+    _key_request_waiters: dict[SessionID, asyncio.Future]
     # Futures that wait for a session to be received (either normally or through a key request)
-    _inbound_session_waiters: Dict[SessionID, asyncio.Future]
+    _inbound_session_waiters: dict[SessionID, asyncio.Future]
 
-    _prev_unwedge: Dict[IdentityKey, float]
+    _prev_unwedge: dict[IdentityKey, float]
+    _fetch_keys_lock: asyncio.Lock
+    _cs_fetch_attempted: set[UserID]
 
     async def wait_for_session(
-        self, room_id: RoomID, sender_key: IdentityKey, session_id: SessionID, timeout: float = 3
+        self, room_id: RoomID, session_id: SessionID, timeout: float = 3
     ) -> bool:
         try:
             fut = self._inbound_session_waiters[session_id]
@@ -62,7 +66,7 @@ class BaseOlmMachine:
         try:
             return await asyncio.wait_for(asyncio.shield(fut), timeout)
         except asyncio.TimeoutError:
-            return await self.crypto_store.has_group_session(room_id, sender_key, session_id)
+            return await self.crypto_store.has_group_session(room_id, session_id)
 
     def _mark_session_received(self, session_id: SessionID) -> None:
         try:
@@ -77,12 +81,13 @@ canonical_json = functools.partial(
 
 
 def verify_signature_json(
-    data: "SignedObject", user_id: UserID, device_id: DeviceID, key: SigningKey
+    data: "SignedObject", user_id: UserID, key_name: DeviceID | str, key: SigningKey
 ) -> bool:
     data_copy = {**data}
     data_copy.pop("unsigned", None)
     signatures = data_copy.pop("signatures")
-    signature = signatures[user_id][f"{EncryptionKeyAlgorithm.ED25519}:{device_id}"]
+    key_id = str(KeyID(EncryptionKeyAlgorithm.ED25519, key_name))
+    signature = signatures[user_id][key_id]
     signed_data = canonical_json(data_copy)
     try:
         olm.ed25519_verify(key, signed_data, signature)
