@@ -16,11 +16,15 @@ from .auth import (
 )
 from .crypto import (
     ClaimKeysResponse,
+    CrossSigner,
+    CrossSigningKeys,
+    CrossSigningUsage,
     DecryptedOlmEvent,
     DeviceIdentity,
     DeviceKeys,
     OlmEventKeys,
     QueryKeysResponse,
+    TOFUSigningKey,
     TrustState,
     UnsignedDeviceInfo,
 )
@@ -34,6 +38,10 @@ from .event import (
     BaseMessageEventContentFuncs,
     BaseRoomEvent,
     BaseUnsigned,
+    BatchSendEvent,
+    BatchSendStateEvent,
+    BeeperMessageStatusEvent,
+    BeeperMessageStatusEventContent,
     CallAnswerEventContent,
     CallCandidate,
     CallCandidatesEventContent,
@@ -68,6 +76,7 @@ from .event import (
     JoinRule,
     JoinRulesStateEventContent,
     JSONWebKey,
+    KeyID,
     KeyRequestAction,
     LocationInfo,
     LocationMessageEventContent,
@@ -77,6 +86,8 @@ from .event import (
     MemberStateEventContent,
     MessageEvent,
     MessageEventContent,
+    MessageStatus,
+    MessageStatusReason,
     MessageType,
     MessageUnsigned,
     OlmCiphertext,
@@ -133,13 +144,13 @@ from .misc import (
     DeviceLists,
     DeviceOTKCount,
     DirectoryPaginationToken,
+    EventContext,
     PaginatedMessages,
     PaginationDirection,
     RoomAliasInfo,
     RoomCreatePreset,
     RoomDirectoryResponse,
     RoomDirectoryVisibility,
-    VersionsResponse,
 )
 from .primitive import (
     JSON,
@@ -152,6 +163,7 @@ from .primitive import (
     RoomAlias,
     RoomID,
     SessionID,
+    Signature,
     SigningKey,
     SyncToken,
     UserID,
@@ -181,6 +193,7 @@ from .util import (
     field,
     serializer,
 )
+from .versions import SpecVersions, Version, VersionFormat, VersionsResponse
 
 __all__ = [
     "DiscoveryInformation",
@@ -198,11 +211,15 @@ __all__ = [
     "UserIdentifierType",
     "WhoamiResponse",
     "ClaimKeysResponse",
+    "CrossSigner",
+    "CrossSigningKeys",
+    "CrossSigningUsage",
     "DecryptedOlmEvent",
     "DeviceIdentity",
     "DeviceKeys",
     "OlmEventKeys",
     "QueryKeysResponse",
+    "TOFUSigningKey",
     "TrustState",
     "UnsignedDeviceInfo",
     "AccountDataEvent",
@@ -214,6 +231,10 @@ __all__ = [
     "BaseMessageEventContentFuncs",
     "BaseRoomEvent",
     "BaseUnsigned",
+    "BatchSendEvent",
+    "BatchSendStateEvent",
+    "BeeperMessageStatusEvent",
+    "BeeperMessageStatusEventContent",
     "CallAnswerEventContent",
     "CallCandidate",
     "CallCandidatesEventContent",
@@ -248,6 +269,7 @@ __all__ = [
     "JoinRule",
     "JoinRulesStateEventContent",
     "JSONWebKey",
+    "KeyID",
     "KeyRequestAction",
     "LocationInfo",
     "LocationMessageEventContent",
@@ -257,6 +279,8 @@ __all__ = [
     "MemberStateEventContent",
     "MessageEvent",
     "MessageEventContent",
+    "MessageStatus",
+    "MessageStatusReason",
     "MessageType",
     "MessageUnsigned",
     "OlmCiphertext",
@@ -322,13 +346,13 @@ __all__ = [
     "DeviceLists",
     "DeviceOTKCount",
     "DirectoryPaginationToken",
+    "EventContext",
     "PaginatedMessages",
     "PaginationDirection",
     "RoomAliasInfo",
     "RoomCreatePreset",
     "RoomDirectoryResponse",
     "RoomDirectoryVisibility",
-    "VersionsResponse",
     "JSON",
     "BatchID",
     "ContentURI",
@@ -339,6 +363,7 @@ __all__ = [
     "RoomAlias",
     "RoomID",
     "SessionID",
+    "Signature",
     "SigningKey",
     "SyncToken",
     "UserID",
@@ -365,4 +390,8 @@ __all__ = [
     "deserializer",
     "field",
     "serializer",
+    "SpecVersions",
+    "Version",
+    "VersionFormat",
+    "VersionsResponse",
 ]
