@@ -10,6 +10,13 @@ from .account_data import (
     RoomTagInfo,
 )
 from .base import BaseEvent, BaseRoomEvent, BaseUnsigned, GenericEvent
+from .batch import BatchSendEvent, BatchSendStateEvent
+from .beeper import (
+    BeeperMessageStatusEvent,
+    BeeperMessageStatusEventContent,
+    MessageStatus,
+    MessageStatusReason,
+)
 from .encrypted import (
     EncryptedEvent,
     EncryptedEventContent,
@@ -17,6 +24,7 @@ from .encrypted import (
     EncryptedOlmEventContent,
     EncryptionAlgorithm,
     EncryptionKeyAlgorithm,
+    KeyID,
     OlmCiphertext,
     OlmMsgType,
 )
