@@ -15,6 +15,11 @@ from .encrypted import EncryptionAlgorithm
 from .type import EventType, RoomType
 
 
+@dataclass
+class NotificationPowerLevels(SerializableAttrs):
+    room: int = 50
+
+
 @dataclass
 class PowerLevelStateEventContent(SerializableAttrs):
     """The content of a power level event."""
@@ -27,6 +32,8 @@ class PowerLevelStateEventContent(SerializableAttrs):
     )
     events_default: int = 0
 
+    notifications: NotificationPowerLevels = attr.ib(factory=lambda: NotificationPowerLevels())
+
     state_default: int = 50
 
     invite: int = 50
